
from .pull_sync1 import (  # noqa: F401
    ModQueueListingPaginator as ModQueueListingPaginator,
    ModQueueSubmissionListingPaginator as ModQueueSubmissionListingPaginator,
    ModQueueCommentListingPaginator as ModQueueCommentListingPaginator,
    ReportsListingPaginator as ReportsListingPaginator,
    ReportsSubmissionListingPaginator as ReportsSubmissionListingPaginator,
    ReportsCommentListingPaginator as ReportsCommentListingPaginator,
    SpamListingPaginator as SpamListingPaginator,
    SpamSubmissionListingPaginator as SpamSubmissionListingPaginator,
    SpamCommentListingPaginator as SpamCommentListingPaginator,
    EditedListingPaginator as EditedListingPaginator,
    EditedSubmissionListingPaginator as EditedSubmissionListingPaginator,
    EditedCommentListingPaginator as EditedCommentListingPaginator,
    UnmoderatedSubmissionListingPaginator as UnmoderatedSubmissionListingPaginator,
)
from .pull_users_sync1 import (  # noqa: F401
    ModeratorsPaginator as ModeratorsPaginator,
    ContributorsPaginator as ContributorsPaginator,
    BannedPaginator as BannedPaginator,
    MutedPaginator as MutedPaginator,
)
from .legacy_pull_users_sync1 import (  # noqa: F401
    UserRelationshipItemListingPaginator as UserRelationshipItemListingPaginator,
    BannedUserRelationshipItemListingPaginator as BannedUserRelationshipItemListingPaginator,
)
from .pull_actions_sync1 import ModerationActionLogPaginator as ModerationActionLogPaginator  # noqa: F401




from __future__ import annotations
from typing import Mapping, Any

from datetime import datetime, timezone

from .artifact import IArtifact

class BaseUser(IArtifact):
    class Subreddit:
        def __init__(self, d: Mapping[str, Any]):
            d = d['subreddit']
            self.name: str = d['display_name']
            self.openness: str = d['subreddit_type']
            self.subscriber_count: int = d['subscribers']
            self.title: str = d['title']
            self.public_description: str = d['public_description']
            self.nsfw: bool = d['over_18']

    class Me:
        def __init__(self, d: Mapping[str, Any]):
            self.is_friend: bool = d['is_friend']
            self.is_blocked: bool = d['is_blocked']

    def __init__(self, d: Mapping[str, Any]):
        self.d: Mapping[str, Any] = d
        self.id36: str = d['id']
        self.id: int = int(self.id36, 36)
        self.created_ut: int = int(d['created_utc'])
        self.created_at: datetime = datetime.fromtimestamp(self.created_ut, timezone.utc)
        self.name: str = d['name']

        self.submission_karma: int = d['link_karma']
        self.comment_karma: int = d['comment_karma']
        self.awardee_karma: int = d.get('awardee_karma', -1)
        self.awarder_karma: int = d.get('awarder_karma', -1)
        self.total_karma: int = d.get('total_karma', -1)

        self.has_premium: bool = d['is_gold']
        self.has_verified_email: bool = d['has_verified_email']

        self.is_admin: bool = d['is_employee']
        self.is_a_subreddit_moderator: bool = d['is_mod']

        self.icon_img: str = d['icon_img']

        self.subreddit: BaseUser.Subreddit = self.Subreddit(d)
        self.me: BaseUser.Me = self.Me(d)


class BaseSuspendedUser(IArtifact):
    class Me:
        def __init__(self, d: Mapping[str, Any]):
            self.is_blocked: bool = d['is_blocked']

    def __init__(self, d: Mapping[str, Any]):
        self.name: str = d['name']
        self.awardee_karma: int = d['awardee_karma']
        self.awarder_karma: int = d['awarder_karma']
        self.total_karma: int = d['total_karma']

        self.me: BaseSuspendedUser.Me = self.Me(d)




from . import (websocket_connection_SYNC, websocket_connection_ASYNC)  # noqa: F401
from . import const  # noqa: F401
from . import exceptions  # noqa: F401
from . import events  # noqa: F401
from . import utils  # noqa: F401
from . import transport  # noqa: F401




from __future__ import annotations
from typing import TYPE_CHECKING, Any, Mapping
if TYPE_CHECKING:
    from ..client_SYNC import Client

from ..models.wiki_base import BaseWikiPageRevisionAuthorUser
from ..models.wiki_SYNC import WikiPageRevisionAuthorUser, WikiPage, WikiPageRevision, WikiPageSettings
from .wiki_base import load_base_wiki_page, load_base_wiki_page_revision, load_base_wiki_page_settings


def promote_base_wiki_page_revision_author_user(base: BaseWikiPageRevisionAuthorUser, client: Client) -> WikiPageRevisionAuthorUser:
    return WikiPageRevisionAuthorUser(base.d, client)

def load_wiki_page_revision_author_user(d: Mapping[str, Any], client: Client) -> WikiPageRevisionAuthorUser:
    return WikiPageRevisionAuthorUser(d, client)


def load_wiki_page(d: Mapping[str, Any], client: Client) -> WikiPage:
    u = load_base_wiki_page(d)
    return WikiPage(
        d=u.d,
        body=u.body,
        body_html=u.body_html,
        can_revise=u.can_revise,
        revision_uuid=u.revision_uuid,
        revision_timestamp=u.revision_timestamp,
        revision_author=promote_base_wiki_page_revision_author_user(u.revision_author, client),
        revision_message=u.revision_message,
    )

def load_wiki_page_revision(d: Mapping[str, Any], client: Client) -> WikiPageRevision:
    u = load_base_wiki_page_revision(d)
    return WikiPageRevision(
        d=u.d,
        uuid=u.uuid,
        timestamp=u.timestamp,
        author=promote_base_wiki_page_revision_author_user(u.author, client),
        message=u.message,
        hidden=u.hidden,
    )


def load_wiki_page_settings(d: Mapping[str, Any], client: Client) -> WikiPageSettings:
    u = load_base_wiki_page_settings(d)
    return WikiPageSettings(
        permlevel=u.permlevel,
        # Type ignore due to https://github.com/python/mypy/issues/10986
        editors=[promote_base_wiki_page_revision_author_user(o, client) for o in u.editors],  # type: ignore[misc]
        unlisted=u.unlisted,
    )




from __future__ import annotations
from typing import Any, Mapping

from ..models.karma_breakdown_entry import KarmaBreakdownEntry

def load_karma_breakdown_entry(d: Mapping[str, Any]) -> KarmaBreakdownEntry:
    return KarmaBreakdownEntry(
        sr_name=d['sr'],
        comment_karma=d['comment_karma'],
        submission_karma=d['link_karma'],
    )




from typing import Any, Mapping

from datetime import datetime
from dataclasses import dataclass

from .artifact import IArtifact

@dataclass(repr=False, eq=False)
class ModerationActionLogEntry(IArtifact):
    #_: KW_ONLY
    d: Mapping[str, Any]
    uuid: str
    timestamp: int
    datetime: datetime
    action: str
    agent_id: int
    agent: str




from __future__ import annotations
from typing import TYPE_CHECKING, Any, Mapping
if TYPE_CHECKING:
    from ..models.moderated_subreddit import ModeratedSubreddit

from ..models.moderated_subreddit import ModeratedUserSubreddit, ModeratedRegularSubreddit

def load_moderated_subreddit(d: Mapping[str, Any]) -> ModeratedSubreddit:
    if d['display_name'].startswith('u_'):
        return ModeratedUserSubreddit(d)
    return ModeratedRegularSubreddit(d)




from __future__ import annotations

from .listing.mixins.subreddit_detail_SYNC import SubredditDetail
from .listing.submission_listing_paginator import SubmissionListingPaginator
from ...models.submission_SYNC import Submission
from .listing.mixins.time_SYNC import Time
from .listing.submission_and_comment_listing_paginator import SubmissionAndExtraSubmissionFieldsCommentListingPaginator

class BestListingPaginator(
    SubredditDetail[Submission],
    SubmissionListingPaginator,
): pass

class HotListingPaginator(
    SubredditDetail[Submission],
    SubmissionListingPaginator,
): pass

class NewListingPaginator(
    SubredditDetail[Submission],
    SubmissionListingPaginator,
): pass

class TopListingPaginator(
    Time[Submission],
    SubredditDetail[Submission],
    SubmissionListingPaginator,
): pass

class RisingListingPaginator(
    SubredditDetail[Submission],
    SubmissionListingPaginator,
): pass

class ControversialListingPaginator(
    Time[Submission],
    SubredditDetail[Submission],
    SubmissionListingPaginator,
): pass

class GildedListingPaginator(
    Time[object],
    SubredditDetail[object],
    SubmissionAndExtraSubmissionFieldsCommentListingPaginator,
): pass




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Sequence
if TYPE_CHECKING:
    from ...client_ASYNC import Client
    from ...models.submission_collection_ASYNC import SubmissionCollectionDetails, SubmissionCollection

from functools import cached_property

from ...model_loaders.submission_collection_ASYNC import load_submission_collection_details, load_submission_collection
from ...util.base_conversion import to_base36
from .create_ASYNC import Create
from .add_post_ASYNC import AddPost
from .remove_post_ASYNC import RemovePost
from .reorder_ASYNC import Reorder

class CollectionProcedures:
    def __init__(self, client: Client) -> None:
        self._client = client
        self.create: Create = Create(client)
        self.add_post: AddPost = AddPost(client)
        self.remove_post: RemovePost = RemovePost(client)
        self.reorder: Reorder = Reorder(client)

    async def get(self, uuid: str) -> Optional[SubmissionCollection]:
        params = {'collection_id': uuid}
        root = await self._client.request('GET', '/api/v1/collections/collection', params=params)
        if len(root) < 3:
            return None
        return load_submission_collection(root, self._client)

    async def get_unfilled(self, uuid: str) -> Optional[SubmissionCollectionDetails]:
        params = {'collection_id': uuid, 'include_links': '0'}
        root = await self._client.request('GET', '/api/v1/collections/collection', params=params)
        if len(root) < 3:
            return None
        return load_submission_collection_details(root, self._client)

    class _get_subreddit_collections_details:
        def __init__(self, outer: CollectionProcedures):
            self._client = outer._client

        async def __call__(self, id: int) -> Sequence[SubmissionCollectionDetails]:
            id36 = to_base36(id)
            return await self.by_id36(id36)

        async def by_id36(self, id36: str) -> Sequence[SubmissionCollectionDetails]:
            params = {'sr_fullname': 't5_' + id36}
            root = await self._client.request('GET', '/api/v1/collections/subreddit_collections', params=params)
            return [load_submission_collection_details(d, self._client) for d in root]

    get_subreddit_collections_details: cached_property[_get_subreddit_collections_details] = \
            cached_property(_get_subreddit_collections_details)

    async def delete(self, uuid: str) -> None:
        params = {'collection_id': uuid}
        await self._client.request('POST', '/api/v1/collections/delete_collection', params=params)

    async def set_title(self, uuid: str, title: str) -> None:
        params = {'collection_id': uuid, 'title': title}
        await self._client.request('POST', '/api/v1/collections/update_collection_title', params=params)

    async def set_description(self, uuid: str, desc: str) -> None:
        params = {'collection_id': uuid, 'description': desc}
        await self._client.request('POST', '/api/v1/collections/update_collection_description', params=params)

    async def set_display_layout(self, uuid: str, layout: Optional[str]) -> None:
        params = {'collection_id': uuid}
        if layout is not None:
            params['display_layout'] = layout
        await self._client.request('POST', '/api/v1/collections/update_collection_display_layout', params=params)

    async def follow(self, uuid: str) -> None:
        params = {'follow': '1'}
        await self._client.request('POST', '/api/v1/collections/follow_collection', params=params)

    async def unfollow(self, uuid: str) -> None:
        params = {'follow': '0'}
        await self._client.request('POST', '/api/v1/collections/follow_collection', params=params)




from __future__ import annotations
from typing import TYPE_CHECKING, Mapping, Any
if TYPE_CHECKING:
    from ...http.requestor_ASYNC import Requestor

from ...http.request import make_request
from ...http.util.json_load import json_loads_response
from .token import Token
from ...auth.utils import apply_basic_auth
from .exceptions import raise_for_token_server_response_error
from .const import TOKEN_OBTAINMENT_URL, REDDIT_MOBILE_IOS_CLIENT_ID, GRANT_DATA

class RedditInternalAPITokenObtainmentClient:
    def __init__(self, requestor: Requestor, uri: str,
            client_id: str,
            grant_data: Any):
        self.requestor: Requestor = requestor
        self.uri: str = uri
        self.client_id: str = client_id
        self.grant_data: Any = grant_data

    async def fetch_data(self) -> Mapping[str, Any]:
        r = make_request('POST', self.uri, json=self.grant_data)
        apply_basic_auth(r, self.client_id, '')
        resp = await self.requestor.send(r)

        try:
            resp_json = json_loads_response(resp)
        except ValueError:
            resp.raise_for_status()
            raise

        raise_for_token_server_response_error(resp_json)
        resp.raise_for_status()
        return resp_json

    async def fetch_token(self) -> Token:
        return Token.from_dict(await self.fetch_data())


def new_reddit_internal_api_token_obtainment_client(requestor: Requestor) -> RedditInternalAPITokenObtainmentClient:
    return RedditInternalAPITokenObtainmentClient(
        requestor,
        uri=TOKEN_OBTAINMENT_URL,
        client_id=REDDIT_MOBILE_IOS_CLIENT_ID,
        grant_data=GRANT_DATA,
    )




from __future__ import annotations
from typing import Mapping, Any, Optional, Sequence, TypeVar, overload, Iterator, Union

from datetime import datetime, timezone

from .artifact import Artifact

class BaseSubmissionCollectionDetails(Artifact):
    def __init__(self, d: Mapping[str, Any]):
        super().__init__(d)
        self.uuid: str = d['collection_id']

        full_id36: str = d['subreddit_id']
        _, _, id36 = full_id36.partition('_')
        self.subreddit_id36: str = id36
        self.subreddit_id: int = int(id36, 36)

        full_id36 = d['author_id']
        _, _, id36 = full_id36.partition('_')
        self.author_id36: str = id36
        self.author_id: int = int(id36, 36)

        self.author_name: str = d['author_name']
        self.title: str = d['title']
        self.description: str = d['description']

        self.created_ts: float = d['created_at_utc']
        self.created_at: datetime = datetime.fromtimestamp(self.created_ts, timezone.utc)
        self.last_post_added_ut: float = d['last_update_utc']
        self.last_post_added_at: datetime = datetime.fromtimestamp(self.last_post_added_ut, timezone.utc)

        self.display_layout: Optional[str] = d['display_layout']
        self.layout: str = 'TIMELINE'
        if self.display_layout is not None:
            self.layout = self.display_layout

        submission_full_id36s: Sequence[str] = d['link_ids']
        self.submission_id36s: Sequence[str] = [s[3:] for s in submission_full_id36s]
        self.submission_ids: Sequence[int] = [int(s, 36) for s in self.submission_id36s]

class BaseSubmissionCollection(BaseSubmissionCollectionDetails):
    def __init__(self, d: Mapping[str, Any]):
        super().__init__(d)

TSubmission = TypeVar('TSubmission')

class GBaseSubmissionCollection(BaseSubmissionCollection, Sequence[TSubmission]):
    def __init__(self, d: Mapping[str, Any]):
        super().__init__(d)
        children_data = d['sorted_links']['data']['children']
        subms = [self._load_submission(i['data']) for i in children_data]
        self.submissions: Sequence[TSubmission] = subms

    def __len__(self) -> int:
        return len(self.submissions)

    def __contains__(self, item: object) -> bool:
        return item in self.submissions

    def __iter__(self) -> Iterator[TSubmission]:
        return iter(self.submissions)

    @overload
    def __getitem__(self, index: int) -> TSubmission: ...
    @overload
    def __getitem__(self, index: slice) -> Sequence[TSubmission]: ...
    def __getitem__(self, index: Union[int, slice]) -> Union[TSubmission, Sequence[TSubmission]]:
        return self.submissions[index]

    def _load_submission(self, m: Mapping[str, Any]) -> TSubmission:
        raise NotImplementedError




from __future__ import annotations
from typing import TYPE_CHECKING, Optional
if TYPE_CHECKING:
    from ...client_SYNC import Client
    from ...models.comment_tree_SYNC import SubmissionTreeNode
    from .SYNC import CommentTreeProcedures

from ...util.base_conversion import to_base36
from ... import http

class Get:
    def __init__(self, outer: CommentTreeProcedures, client: Client):
        self._outer = outer
        self._client = client

    def __call__(self,
        submission_id: int,
        comment_id: Optional[int] = None,
        *,
        sort: str = 'confidence',
        limit: Optional[int] = None,
        depth: Optional[int] = None,
        context: Optional[int] = None,
    ) -> Optional[SubmissionTreeNode]:
        submission_id36 = to_base36(submission_id)
        comment_id36 = comment_id if comment_id is None else to_base36(comment_id)
        return self.by_id36(submission_id36, comment_id36, sort=sort, limit=limit, depth=depth, context=context)

    def by_id36(self,
        submission_id36: str,
        comment_id36: Optional[str] = None,
        *,
        sort: str = 'confidence',
        limit: Optional[int] = None,
        depth: Optional[int] = None,
        context: Optional[int] = None,
    ) -> Optional[SubmissionTreeNode]:
        try:
            return self._outer.fetch.by_id36(
                submission_id36,
                comment_id36,
                sort=sort,
                limit=limit,
                depth=depth,
                context=context,
            )
        except http.exceptions.StatusCodeException as e:
            if e.status_code == 404:
                return None
            raise




from . import (SYNC, ASYNC)  # noqa: F401
from . import (session_base_SYNC, session_base_ASYNC)  # noqa: F401
from . import (http_client_SYNC, http_client_ASYNC)  # noqa: F401
from . import request  # noqa: F401
from .request import Request as Request, make_request as make_request  # noqa: F401
from . import response  # noqa: F401
from .response import Response as Response  # noqa: F401
from . import exceptions  # noqa: F401
from . import payload  # noqa: F401
from . import misc  # noqa: F401
from . import transport  # noqa: F401
from . import util  # noqa: F401




from __future__ import annotations
from typing import TypeVar, Iterator, Optional, overload, Callable, Union
from collections.abc import Collection

T = TypeVar('T')

class EventDispatcher(Collection[T]):
    def __init__(self) -> None:
        self._handlers: set[T] = set()

    def __len__(self) -> int:
        return len(self._handlers)

    def __contains__(self, item: object) -> bool:
        return item in self._handlers

    def __iter__(self) -> Iterator[T]:
        return iter(self._handlers)

    def attach(self, handler: T) -> None:
        self._handlers.add(handler)

    def detach(self, handler: T) -> None:
        self._handlers.discard(handler)

    @overload
    def attach_handler(self, handler: T) -> T: ...
    @overload
    def attach_handler(self) -> Callable[[T], T]: ...
    def attach_handler(self, handler: Optional[T] = None) -> Union[T, Callable[[T], T]]:
        if handler is None:
            return self.attach_handler
        self.attach(handler)
        return handler




from typing import Protocol

def _main() -> None:
    class Example:
        pass

    class ExampleHandlerProtocol(Protocol):
        def __call__(self, example: Example, /) -> None:
            ...

    class ExampleEventDispatcher(EventDispatcher[ExampleHandlerProtocol]):
        def __call__(self, example: Example) -> None:
            for handler in self:
                handler(example)

    def handler1(example: Example) -> None:
        print('Got:', example)

    example_event = ExampleEventDispatcher()
    example_event.attach(handler1)

    @example_event.attach_handler
    def handler2(obj: object) -> None:
        print('Got:', obj)

    @example_event.attach_handler()
    def handler3(obj: object) -> None:
        print('Got:', obj)

    example_event(Example())

if __name__ == '__main__':
    _main()




from __future__ import annotations
from typing import Sequence

from .listing.listing_paginator import ListingPaginator
from ...models.message_SYNC import MailboxMessage, ComposedMessage, CommentMessage
from ...model_loaders.message_SYNC import (
    load_message,
    load_composed_message,
    load_comment_message,
    load_composed_message_thread,
)

class MessageListingPaginator(ListingPaginator[MailboxMessage]):
    def fetch(self) -> Sequence[MailboxMessage]:
        data = self._fetch_data()
        return [load_message(d['data'], self.client) for d in data['children']]

class ComposedMessageListingPaginator(ListingPaginator[ComposedMessage]):
    def fetch(self) -> Sequence[ComposedMessage]:
        data = self._fetch_data()
        return [load_composed_message(d['data'], self.client) for d in data['children']]

class CommentMessageListingPaginator(ListingPaginator[CommentMessage]):
    def fetch(self) -> Sequence[CommentMessage]:
        data = self._fetch_data()
        return [load_comment_message(d['data'], self.client) for d in data['children']]

class ComposedMessageThreadListingPaginator(ListingPaginator[Sequence[ComposedMessage]]):
    def fetch(self) -> Sequence[Sequence[ComposedMessage]]:
        data = self._fetch_data()
        return [load_composed_message_thread(d['data'], self.client) for d in data['children']]



#!/usr/bin/env python3
"""
Exchange an authorization code for an access/refresh token.
"""

from __future__ import annotations
from typing import TYPE_CHECKING, Optional
if TYPE_CHECKING:
    from types import FrameType


import argparse
class Formatter(argparse.ArgumentDefaultsHelpFormatter, argparse.RawDescriptionHelpFormatter): pass
parser = argparse.ArgumentParser(description=__doc__, formatter_class=Formatter)
parser._optionals.title = __import__('gettext').gettext('named arguments')
add = parser.add_argument
add('client_id', nargs='?')
add('client_secret', nargs='?')
add('authorization_code', nargs='?')
add('--client-id', metavar='CLIENT_ID', dest='client_id_opt', help=argparse.SUPPRESS)
add('--client-secret', metavar='CLIENT_SECRET', dest='client_secret_opt', help=argparse.SUPPRESS)
add('--authorization-code', metavar='CLIENT_SECRET', dest='authorization_code_opt', help=argparse.SUPPRESS)
add('--redirect-uri', default="http://localhost:8080", help="\N{ZERO WIDTH SPACE}")
args = parser.parse_args()


import sys
import os
import signal
from functools import partial
from pprint import pp

import redditwarp
from redditwarp.http.transport.SYNC import load_transport, new_session
from redditwarp.core.SYNC import RedditTokenObtainmentClient
from redditwarp.util.user_agent_SYNC import get_user_agent_from_session

def get_client_cred_input(prompt: str, env: str, v: Optional[str]) -> str:
    if v is None:
        v = input(prompt)
    if v == '.':
        v = os.environ[env]
        print(v)
    return v

def get_client_id(args: argparse.Namespace) -> str:
    v = args.client_id_opt or args.client_id
    return get_client_cred_input('Client ID: ', 'redditwarp_client_id', v)

def get_client_secret(args: argparse.Namespace) -> str:
    v = args.client_secret_opt or args.client_secret
    return get_client_cred_input('Client secret: ', 'redditwarp_client_secret', v)

if not sys.flags.interactive:
    @partial(signal.signal, signal.SIGINT)
    def handle_sigint(sig: int, frame: Optional[FrameType]) -> None:
        print('KeyboardInterrupt', file=sys.stderr)
        sys.exit(130)

load_transport()

client_id = get_client_id(args)
client_secret = get_client_secret(args)
code: str = args.authorization_code_opt or args.authorization_code or input('Code: ')
redirect_uri: str = args.redirect_uri

print('* Step 3. Exchange the authorization code for an access/refresh token.\n')

grant = redditwarp.auth.grants.AuthorizationCodeGrant(code, redirect_uri)
print('Authorization grant:')
pp(dict(grant))
print()

session = new_session()
ua = get_user_agent_from_session(session) + " redditwarp.cli.exchange_authorization_code"
token_client = RedditTokenObtainmentClient(
    session,
    redditwarp.auth.const.TOKEN_OBTAINMENT_URL,
    (client_id, client_secret),
    grant,
    headers={'User-Agent': ua},
)

print('Obtaining token(s) from token server...\n')

token = token_client.fetch_token()

print(f'''\
     Access token: {token.access_token}
    Refresh token: {token.refresh_token}''')




from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ..client_ASYNC import Client

from .submission.ASYNC import SubmissionProcedures
from .comment.ASYNC import CommentProcedures
from .subreddit.ASYNC import SubredditProcedures
from .comment_tree.ASYNC import CommentTreeProcedures
from .front.ASYNC import FrontProcedures
from .user.ASYNC import UserProcedures
from .account.ASYNC import AccountProcedures
from .collection.ASYNC import CollectionProcedures
from .flair.ASYNC import FlairProcedures
from .flair_emoji.ASYNC import FlairEmojiProcedures
from .custom_feed.ASYNC import CustomFeedProcedures
from .live_thread.ASYNC import LiveThreadProcedures
from .message.ASYNC import MessageProcedures
from .moderation.ASYNC import ModerationProcedures
from .draft.ASYNC import DraftProcedures
from .misc.ASYNC import MiscProcedures
from .modmail.ASYNC import ModmailProcedures
from .widget.ASYNC import WidgetProcedures
from .wiki.ASYNC import WikiProcedures
from .legacy_reddit_subreddit_style.ASYNC import LegacyRedditSubredditStyleProcedures
from .redesign_reddit_subreddit_style.ASYNC import RedesignRedditSubredditStyleProcedures

class SiteProcedures:
    def __init__(self, client: Client):
        self._client = client
        self.submission: SubmissionProcedures = SubmissionProcedures(client)
        self.comment: CommentProcedures = CommentProcedures(client)
        self.subreddit: SubredditProcedures = SubredditProcedures(client)
        self.comment_tree: CommentTreeProcedures = CommentTreeProcedures(client)
        self.front: FrontProcedures = FrontProcedures(client)
        self.user: UserProcedures = UserProcedures(client)
        self.account: AccountProcedures = AccountProcedures(client)
        self.collection: CollectionProcedures = CollectionProcedures(client)
        self.flair: FlairProcedures = FlairProcedures(client)
        self.flair_emoji: FlairEmojiProcedures = FlairEmojiProcedures(client)
        self.custom_feed: CustomFeedProcedures = CustomFeedProcedures(client)
        self.live_thread: LiveThreadProcedures = LiveThreadProcedures(client)
        self.message: MessageProcedures = MessageProcedures(client)
        self.moderation: ModerationProcedures = ModerationProcedures(client)
        self.draft: DraftProcedures = DraftProcedures(client)
        self.misc: MiscProcedures = MiscProcedures(client)
        self.modmail: ModmailProcedures = ModmailProcedures(client)
        self.widget: WidgetProcedures = WidgetProcedures(client)
        self.wiki: WikiProcedures = WikiProcedures(client)
        self.legacy_reddit_subreddit_style: LegacyRedditSubredditStyleProcedures = LegacyRedditSubredditStyleProcedures(client)
        self.redesign_reddit_subreddit_style: RedesignRedditSubredditStyleProcedures = RedesignRedditSubredditStyleProcedures(client)




from __future__ import annotations
from typing import TYPE_CHECKING, Iterable, Optional
if TYPE_CHECKING:
    from ...client_SYNC import Client
    from ...models.comment_tree_SYNC import MoreCommentsTreeNode

from functools import cached_property

from ...util.base_conversion import to_base36
from ...model_loaders.comment_tree_SYNC import load_more_children
from .get_SYNC import Get
from .fetch_SYNC import Fetch

class CommentTreeProcedures:
    def __init__(self, client: Client):
        self._client = client
        self.get: Get = Get(self, client)
        self.fetch: Fetch = Fetch(self, client)

    class _more_children:
        def __init__(self, outer: CommentTreeProcedures):
            self._outer = outer
            self._client = outer._client

        def __call__(self,
            submission_id: int,
            child_ids: Iterable[int],
            *,
            sort: str = '',
            depth: Optional[int] = None,
            exact: bool = False,
        ) -> MoreCommentsTreeNode:
            return self.by_id36(
                to_base36(submission_id),
                (to_base36(x) for x in child_ids),
                sort=sort,
                depth=depth,
                exact=exact,
            )

        def by_id36(self,
            submission_id36: str,
            child_id36s: Iterable[str],
            *,
            sort: str = '',
            depth: Optional[int] = None,
            exact: bool = False,
        ) -> MoreCommentsTreeNode:
            def g() -> Iterable[tuple[str, str]]:
                yield ('link_id', 't3_' + submission_id36)
                yield ('children', ','.join(child_id36s))
                if sort:
                    yield ('sort', sort)
                if depth is not None:
                    yield ('depth', str(depth))
                if exact:
                    yield ('limit_children', '1')

            resp_data = self._client.request('GET', '/api/morechildren', params=dict(g()))
            return load_more_children(resp_data, self._client, submission_id36, sort)

    more_children: cached_property[_more_children] = cached_property(_more_children)




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, MutableSequence
if TYPE_CHECKING:
    from ..http.request import Request
    from ..http.response import Response
    from ..http.requestor_ASYNC import Requestor

from collections import deque

from ..http.requestor_augmenter_ASYNC import RequestorAugmenter

class Recorded(RequestorAugmenter):
    def __init__(self, requestor: Requestor) -> None:
        super().__init__(requestor)
        self.last_request: Optional[Request] = None
        self.last_response: Optional[Response] = None
        self.last_transfer: Optional[tuple[Request, Response]] = None
        self.last_transmit: Optional[tuple[Request, Optional[Response]]] = None
        self.last_request_queue: MutableSequence[Request] = deque(maxlen=16)
        self.last_response_queue: MutableSequence[Response] = deque(maxlen=16)
        self.last_transfer_queue: MutableSequence[tuple[Request, Response]] = deque(maxlen=16)
        self.last_transmit_queue: MutableSequence[tuple[Request, Optional[Response]]] = deque(maxlen=16)

    async def send(self, request: Request, *, timeout: float = -2) -> Response:
        resp = None
        try:
            resp = await self.requestor.send(request, timeout=timeout)
        finally:
            self.last_request = request
            self.last_transmit = (request, resp)
            self.last_request_queue.append(self.last_request)
            self.last_transmit_queue.append(self.last_transmit)
            if resp is None:
                self.last_response = None
                self.last_transfer = None
            else:
                self.last_response = resp
                self.last_transfer = (request, resp)
                self.last_response_queue.append(self.last_response)
                self.last_transfer_queue.append(self.last_transfer)

        if resp is None:
            raise Exception
        return resp



class Last:
    @property
    def request(self) -> Optional[Request]:
        return self.recorder.last_request
    @property
    def response(self) -> Optional[Response]:
        return self.recorder.last_response
    @property
    def transfer(self) -> Optional[tuple[Request, Response]]:
        return self.recorder.last_transfer
    @property
    def transmit(self) -> Optional[tuple[Request, Optional[Response]]]:
        return self.recorder.last_transmit
    @property
    def request_queue(self) -> MutableSequence[Request]:
        return self.recorder.last_request_queue
    @property
    def response_queue(self) -> MutableSequence[Response]:
        return self.recorder.last_response_queue
    @property
    def transfer_queue(self) -> MutableSequence[tuple[Request, Response]]:
        return self.recorder.last_transfer_queue
    @property
    def transmit_queue(self) -> MutableSequence[tuple[Request, Optional[Response]]]:
        return self.recorder.last_transmit_queue

    def __init__(self, recorder: Recorded) -> None:
        self.recorder: Recorded = recorder




from __future__ import annotations
from typing import TYPE_CHECKING, Any, Callable
if TYPE_CHECKING:
    from ..response import Response

import json

json_decoder: json.JSONDecoder = json.JSONDecoder()
json_decode: Callable[[str], Any] = json_decoder.decode

def json_loads_response(response: Response, /) -> Any:
    if not response.headers.get('Content-Type', '').startswith('application/json'):
        raise ValueError("response Content-Type is not 'application/json'")
    text = response.data.decode()
    return json_decode(text)




from __future__ import annotations
from typing import TYPE_CHECKING, IO, Optional
if TYPE_CHECKING:
    from ...client_ASYNC import Client

from functools import cached_property

from ...models.flair_emoji import FlairEmojiUploadLease, SubredditFlairEmojis
from ...model_loaders.flair_emoji import load_flair_emoji_upload_lease, load_flair_emoji
from ...http.payload import guess_mimetype_from_filename

class FlairEmojiProcedures:
    def __init__(self, client: Client):
        self._client = client

    async def get_subreddit_emojis(self, sr: str) -> SubredditFlairEmojis:
        root = await self._client.request('GET', f'/api/v1/{sr}/emojis/all')
        root = dict(root)
        reddit_emojis_root = root.pop('snoomojis')
        _full_id36: str
        _full_id36, subreddit_emojis_root = root.popitem()
        _, _, subreddit_id36 = _full_id36.partition('_')
        return SubredditFlairEmojis(
            [load_flair_emoji(d, name) for name, d in subreddit_emojis_root.items()],
            [load_flair_emoji(d, name) for name, d in reddit_emojis_root.items()],
            subreddit_id36,
        )

    class _create:
        def __init__(self, outer: FlairEmojiProcedures) -> None:
            self._client = outer._client

        async def __call__(self,
            sr: str,
            name: str,
            file: IO[bytes],
            *,
            post_enabled: bool = True,
            user_enabled: bool = True,
            mod_only: bool = False,
        ) -> None:
            upload_lease = await self.upload(file, sr=sr)
            await self.add(sr, upload_lease.fields['key'], name,
                    post_enabled=post_enabled,
                    user_enabled=user_enabled,
                    mod_only=mod_only)

        async def obtain_upload_lease(self, *, sr: str, filename: str, mimetype: Optional[str] = None) -> FlairEmojiUploadLease:
            if mimetype is None:
                mimetype = guess_mimetype_from_filename(filename)
            result = await self._client.request('POST', f'/api/v1/{sr}/emoji_asset_upload_s3',
                    data={'filepath': filename, 'mimetype': mimetype})
            return load_flair_emoji_upload_lease(result)

        async def deposit_file(self, file: IO[bytes], upload_lease: FlairEmojiUploadLease, *,
                timeout: float = 1000) -> None:
            session = self._client.http.session
            resp = await session.request('POST', upload_lease.endpoint, data=upload_lease.fields,
                    files={'file': file}, timeout=timeout)
            resp.raise_for_status()

        async def upload(self, file: IO[bytes], *, sr: str, timeout: float = 1000) -> FlairEmojiUploadLease:
            upload_lease = await self.obtain_upload_lease(filename=file.name, sr=sr)
            await self.deposit_file(file, upload_lease, timeout=timeout)
            return upload_lease

        async def add(self, sr: str, s3_object_key: str, name: str, *,
                post_enabled: bool = True,
                user_enabled: bool = True,
                mod_only: bool = False) -> None:
            data = {
                's3_key': s3_object_key,
                'name': name,
                'post_flair_allowed': '01'[post_enabled],
                'user_flair_allowed': '01'[user_enabled],
                'mod_flair_only': '01'[mod_only],
            }
            await self._client.request('POST', f'/api/v1/{sr}/emoji', data=data)

    create: cached_property[_create] = cached_property(_create)

    async def set_emoji_permissions(self,
        sr: str,
        emoji_name: str,
        *,
        mod_only: bool = False,
        user_enabled: bool = True,
        post_enabled: bool = True,
    ) -> None:
        data = {
            'name': emoji_name,
            'mod_flair_only': '01'[mod_only],
            'user_flair_allowed': '01'[user_enabled],
            'post_flair_allowed': '01'[post_enabled],
        }
        await self._client.request('POST', f'/api/v1/{sr}/emoji_permissions', data=data)

    async def delete(self, sr: str, emoji_name: str) -> None:
        await self._client.request('DELETE', f'/api/v1/{sr}/emoji/{emoji_name}')

    async def set_custom_emoji_size(self, sr: str, size: Optional[tuple[int, int]]) -> None:
        data = None
        if size is not None:
            w, h = size
            data = {
                'width': str(w),
                'height': str(h),
            }
        await self._client.request('POST', f'/api/v1/{sr}/emoji_custom_size', data=data)

    async def enable_emojis_in_sr(self, sr: str) -> None:
        data = {'subreddit': sr, 'enable': '1'}
        await self._client.request('POST', '/api/enable_emojis_in_sr', data=data)

    async def disable_emojis_in_sr(self, sr: str) -> None:
        data = {'subreddit': sr, 'enable': '0'}
        await self._client.request('POST', '/api/enable_emojis_in_sr', data=data)




from __future__ import annotations
from typing import TYPE_CHECKING, Mapping, Any
if TYPE_CHECKING:
    from ...client_SYNC import Client

import json

class MiscProcedures:
    def __init__(self, client: Client):
        self._client = client

    def convert_rtjson_to_markdown(self, rtjson: Mapping[str, Any]) -> str:
        root = self._client.request('POST', '/api/convert_rte_body_format',
                data={'output_mode': 'markdown', 'richtext_json': json.dumps(rtjson)})
        return root['output']

    def convert_markdown_to_rtjson(self, md: str) -> Mapping[str, Any]:
        root = self._client.request('POST', '/api/convert_rte_body_format',
                data={'output_mode': 'rtjson', 'markdown_text': md})
        return root['output']




from __future__ import annotations

from dataclasses import dataclass

@dataclass
class ModReport:
    reason: str
    agent_name: str

@dataclass
class UserReport:
    reason: str
    count: int
    snoozed: bool
    can_snooze: bool




from .token_obtainment_client_SYNC import TokenObtainmentClient as TokenObtainmentClient  # noqa: F401
from .token_revocation_client_SYNC import TokenRevocationClient as TokenRevocationClient  # noqa: F401




from .async_call_chunk import AsyncCallChunk as AsyncCallChunk  # noqa: F401
from .call_chunk_chaining_async_iterator import CallChunkChainingAsyncIterator as CallChunkChainingAsyncIterator  # noqa: F401
from .call_chunk_calling_async_iterator import CallChunkCallingAsyncIterator as CallChunkCallingAsyncIterator  # noqa: F401




from __future__ import annotations
from typing import TYPE_CHECKING, Mapping, Optional
if TYPE_CHECKING:
    from typing import Iterator

from typing import ClassVar
from dataclasses import dataclass, field

@dataclass(frozen=True)
class AuthorizationGrant(Mapping[str, str]):
    """An authorization grant is a credential representing the resource
    owner's authorization that's used to exchange for a bearer token.

    An empty string should be treated the same as `None` in all fields
    annotated as `Optional`. The annotation is a reflection of the field
    value requirements of various grants types defined in the OAuth2 spec.
    """
    GRANT_TYPE: ClassVar[str] = ''
    _d: dict[str, str] = field(init=False, repr=False, default_factory=dict)

    def __post_init__(self) -> None:
        self._d['grant_type'] = self.GRANT_TYPE
        self._d.update({
            k: v
            for k, v in vars(self).items()
            if not k.startswith('_')
            if v
        })

    def __contains__(self, item: object) -> bool:
        return item in self._d
    def __iter__(self) -> Iterator[str]:
        return iter(self._d)
    def __len__(self) -> int:
        return len(self._d)
    def __getitem__(self, key: str) -> str:
        return self._d[key]

@dataclass(frozen=True)
class AuthorizationCodeGrant(AuthorizationGrant):
    GRANT_TYPE: ClassVar[str] = 'authorization_code'
    code: str
    redirect_uri: Optional[str]
    client_id: Optional[str] = None

@dataclass(frozen=True)
class ResourceOwnerPasswordCredentialsGrant(AuthorizationGrant):
    GRANT_TYPE: ClassVar[str] = 'password'
    username: str
    password: str
    scope: Optional[str] = None

@dataclass(frozen=True)
class ClientCredentialsGrant(AuthorizationGrant):
    GRANT_TYPE: ClassVar[str] = 'client_credentials'
    scope: Optional[str] = None

@dataclass(frozen=True)
class RefreshTokenGrant(AuthorizationGrant):
    GRANT_TYPE: ClassVar[str] = 'refresh_token'
    refresh_token: str
    scope: Optional[str] = None

@dataclass(frozen=True)
class InstalledClientGrant(AuthorizationGrant):
    GRANT_TYPE: ClassVar[str] = "https://oauth.reddit.com/grants/installed_client"
    device_id: str
    scope: Optional[str] = None




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Sequence, Iterable
if TYPE_CHECKING:
    from ...client_ASYNC import Client
    from ...models.comment_ASYNC import (
        Comment,
        EditPostTextEndpointComment,
    )

from ...model_loaders.comment_ASYNC import load_comment, load_edit_post_text_endpoint_comment
from ...util.base_conversion import to_base36
from ...iterators.chunking import chunked
from ...iterators.call_chunk_chaining_async_iterator import CallChunkChainingAsyncIterator
from ...iterators.async_call_chunk import AsyncCallChunk
from .fetch_ASYNC import Fetch
from .get_ASYNC import Get

class CommentProcedures:
    def __init__(self, client: Client):
        self._client = client
        self.fetch: Fetch = Fetch(self, client)
        self.get: Get = Get(client)

    def bulk_fetch(self, ids: Iterable[int]) -> CallChunkChainingAsyncIterator[Comment]:
        async def mass_fetch(ids: Sequence[int]) -> Sequence[Comment]:
            id36s = map(to_base36, ids)
            full_id36s = map('t1_'.__add__, id36s)
            ids_str = ','.join(full_id36s)
            root = await self._client.request('GET', '/api/info', params={'id': ids_str})
            return [load_comment(i['data'], self._client) for i in root['data']['children']]

        return CallChunkChainingAsyncIterator(AsyncCallChunk(mass_fetch, idfs) for idfs in chunked(ids, 100))

    async def reply(self, comment_id: int, text: str) -> Comment:
        data = {
            'thing_id': 't1_' + to_base36(comment_id),
            'text': text,
            'return_rtjson': '1',
        }
        result = await self._client.request('POST', '/api/comment', data=data)
        return load_comment(result, self._client)

    async def edit_body(self, comment_id: int, text: str) -> EditPostTextEndpointComment:
        data = {
            'thing_id': 't1_' + to_base36(comment_id),
            'text': text,
            'return_rtjson': '1',
        }
        result = await self._client.request('POST', '/api/editusertext', data=data)
        return load_edit_post_text_endpoint_comment(result, self._client)

    async def delete(self, comment_id: int) -> None:
        data = {'id': 't1_' + to_base36(comment_id)}
        await self._client.request('POST', '/api/del', data=data)

    async def lock(self, comment_id: int) -> None:
        data = {'id': 't1_' + to_base36(comment_id)}
        await self._client.request('POST', '/api/lock', data=data)

    async def unlock(self, comment_id: int) -> None:
        data = {'id': 't1_' + to_base36(comment_id)}
        await self._client.request('POST', '/api/unlock', data=data)

    async def vote(self, comment_id: int, direction: int) -> None:
        data = {
            'id': 't1_' + to_base36(comment_id),
            'dir': str(direction),
        }
        await self._client.request('POST', '/api/vote', data=data)

    async def save(self, comment_id: int, category: Optional[str] = None) -> None:
        data = {
            'id': 't1_' + to_base36(comment_id),
        }
        if category is not None:
            data['category'] = category
        await self._client.request('POST', '/api/save', data=data)

    async def unsave(self, comment_id: int) -> None:
        data = {'id': 't1_' + to_base36(comment_id)}
        await self._client.request('POST', '/api/unsave', data=data)

    async def distinguish(self, comment_id: int) -> Comment:
        data = {
            'id': 't1_' + to_base36(comment_id),
            'how': 'yes',
        }
        root = await self._client.request('POST', '/api/distinguish', data=data)
        return load_comment(root['json']['data']['things'][0]['data'], self._client)

    async def distinguish_and_sticky(self, comment_id: int) -> Comment:
        data = {
            'id': 't1_' + to_base36(comment_id),
            'how': 'yes',
            'sticky': '1',
        }
        root = await self._client.request('POST', '/api/distinguish', data=data)
        return load_comment(root['json']['data']['things'][0]['data'], self._client)

    async def undistinguish(self, comment_id: int) -> Comment:
        data = {
            'id': 't1_' + to_base36(comment_id),
            'how': 'no',
        }
        root = await self._client.request('POST', '/api/distinguish', data=data)
        return load_comment(root['json']['data']['things'][0]['data'], self._client)

    async def enable_reply_notifications(self, comment_id: int) -> None:
        data = {
            'id': 't1_' + to_base36(comment_id),
            'state': '1'
        }
        await self._client.request('POST', '/api/sendreplies', data=data)

    async def disable_reply_notifications(self, comment_id: int) -> None:
        data = {
            'id': 't1_' + to_base36(comment_id),
            'state': '0'
        }
        await self._client.request('POST', '/api/sendreplies', data=data)

    async def approve(self, comment_id: int) -> None:
        data = {'id': 't1_' + to_base36(comment_id)}
        await self._client.request('POST', '/api/approve', data=data)

    async def remove(self, comment_id: int) -> None:
        data = {
            'id': 't1_' + to_base36(comment_id),
            'spam': '0',
        }
        await self._client.request('POST', '/api/remove', data=data)

    async def remove_spam(self, comment_id: int) -> None:
        data = {
            'id': 't1_' + to_base36(comment_id),
            'spam': '1',
        }
        await self._client.request('POST', '/api/remove', data=data)

    async def ignore_reports(self, idn: int) -> None:
        await self._client.request('POST', '/api/ignore_reports', data={'id': 't1_' + to_base36(idn)})

    async def unignore_reports(self, idn: int) -> None:
        await self._client.request('POST', '/api/unignore_reports', data={'id': 't1_' + to_base36(idn)})

    async def snooze_reports(self, idn: int, reason: str) -> None:
        data = {'id': 't1_' + to_base36(idn), 'reason': reason}
        await self._client.request('POST', '/api/snooze_reports', data=data)

    async def unsnooze_reports(self, idn: int, reason: str) -> None:
        data = {'id': 't1_' + to_base36(idn), 'reason': reason}
        await self._client.request('POST', '/api/unsnooze_reports', data=data)

    async def apply_removal_reason(self,
            comment_id: int,
            reason_id: Optional[str],
            note: Optional[str] = None) -> None:
        target = 't1_' + to_base36(comment_id)
        json_data = {'item_ids': [target], 'reason_id': reason_id, 'mod_note': note}
        await self._client.request('POST', '/api/v1/modactions/removal_reasons', json=json_data)

    async def send_removal_comment(self,
            comment_id: int,
            title: str,
            message: str) -> Comment:
        target = 't1_' + to_base36(comment_id)
        json_data = {
            'type': 'public',
            'item_id': [target],
            'title': title,
            'message': message,
        }
        root = await self._client.request('POST', '/api/v1/modactions/removal_comment_message', json=json_data)
        return load_comment(root, self._client)

    async def send_removal_message(self,
            comment_id: int,
            title: str,
            message: str,
            *,
            exposed: bool = False) -> None:
        target = 't1_' + to_base36(comment_id)
        json_data = {
            'type': 'private' + ('_exposed' if exposed else ''),
            'item_id': [target],
            'title': title,
            'message': message,
        }
        await self._client.request('POST', '/api/v1/modactions/removal_comment_message', json=json_data)
